﻿/**
 * @license Copyright © 2013 onwards, Andrew Whewell
 * All rights reserved.
 *
 * Redistribution and use of this software in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 *    * Neither the name of the author nor the names of the program's contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OF THE SOFTWARE BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/**
 * @fileoverview Configuration UI abstractions.
 */

(function(VRS, $, /** object= */ undefined)
{
    //region OptionField
    /**
     * Describes a single editable option - there is usually one of these for each configurable property on an object.
     * @param {object}                      settings
     * @param {string}                      settings.name           The unique name of the field.
     * @param {string}                      settings.controlType    The VRS.optionsControlType enum value that determines which control will be used to display and/or edit this field.
     * @param {string|function():string}   [settings.labelKey]      The index into VRS.$$ of the label to use when displaying the field or a function that returns the translated text.
     * @param {function():*}               [settings.getValue]      A callback that returns the value to display / edit from the object being configured.
     * @param {function(*)}                [settings.setValue]      A callback that is passed the edited value. It is expected to copy the value into the object being configured.
     * @param {function()}                 [settings.saveState]     A callback that can save the current state of the object being configured.
     * @param {object}                     [settings.typeSettings]  A configuration object appropriate for the control type being used.
     * @param {bool}                       [settings.keepWithNext]  If true then the field is to be displayed next to the field that follows it.
     * @param {function(function, *):*}    [settings.hookChanged]   A hook function that can be used to pick up changes to the value that is returned by getValue - for fields whose value may change while the configuration UI is on display.
     * @param {function(*)}                [settings.unhookChanged] An unhook function that can unhook the event hooked by hookChanged.
     * @param {VRS.InputWidth=}            [settings.inputWidth]    An indication of how wide the input field needs to be.
     * @constructor
     */
    VRS.OptionField = function(settings)
    {
        if(!settings) throw 'You must supply settings';
        if(!settings.name) throw 'You must supply a name for the field';
        if(!settings.controlType) throw 'You must supply the field\'s control type';
        if(!VRS.optionControlTypeBroker.controlTypeHasHandler(settings.controlType)) throw 'There is no control type handler for ' + settings.controlType;

        //region -- Fields
        var that = this;
        var _ChangedHookResult = undefined;
        //endregion

        //region -- Properties
        /** @type {string} @private */
        var _Name = settings.name;
        this.getName = function()   { return _Name; };

        /** @type {VRS.optionControlTypes} @private */
        var _ControlType = settings.controlType;
        this.getControlType = function()        { return _ControlType; };
        //noinspection JSUnusedGlobalSymbols
        this.setControlType = function(/** VRS.optionControlTypes */ value)   { _ControlType = value; };

        /** @type {string} @private */
        var _LabelKey = settings.labelKey || '';
        this.getLabelKey = function()       { return _LabelKey; };
        //noinspection JSUnusedGlobalSymbols
        this.setLabelKey = function(/** string */ value)  { _LabelKey = value; };

        /** @type {function():*} @private */
        var _GetValue = settings.getValue || $.noop;
        this.getGetValue = function()       { return _GetValue; };
        //noinspection JSUnusedGlobalSymbols
        this.setGetValue = function(/** function():* */ value)  { _GetValue = value; };

        /** @type {function(*)} @private */
        var _SetValue = settings.setValue || $.noop;
        this.getSetValue = function()       { return _SetValue; };
        //noinspection JSUnusedGlobalSymbols
        this.setSetValue = function(/** function(*) */ value)  { _SetValue = value; };

        /** @type {function()} @private */
        var _SaveState = settings.saveState || $.noop;
        this.getSaveState = function()      { return _SaveState; };
        //noinspection JSUnusedGlobalSymbols
        this.setSaveState = function(/** function() */ value) { _SaveState = value; };

        /** @type {Object} @private */
        var _TypeSettings = settings.typeSettings;
        this.getTypeSettings = function()       { return _TypeSettings; };
        //noinspection JSUnusedGlobalSymbols
        this.setTypeSettings = function(/** object */ value)  { _TypeSettings = value; };

        /** @type {boolean} @private */
        var _KeepWithNext = !!settings.keepWithNext;
        this.getKeepWithNext = function()       { return _KeepWithNext; };
        this.setKeepWithNext = function(/** boolean */ value)  { _KeepWithNext = value; };

        /** @type {function(Function,*):*} @private */
        var _HookChanged = settings.hookChanged;
        //noinspection JSUnusedGlobalSymbols
        this.getHookChanged = function() { return _HookChanged; };
        //noinspection JSUnusedGlobalSymbols
        this.setHookChanged = function(/** function(Function,*):* */ value) { _HookChanged = value; };

        /** @type {function(*)} @private */
        var _Unhook = settings.unhookChanged;
        //noinspection JSUnusedGlobalSymbols
        this.getUnhook = function() { return _Unhook; };
        //noinspection JSUnusedGlobalSymbols
        this.setUnhook = function(/** function(*) */ value) { _Unhook = value; };

        /** @type {VRS.InputWidth} @private */
        var _InputWidth = settings.inputWidth === undefined ? VRS.InputWidth.Auto : settings.inputWidth;
        //noinspection JSUnusedGlobalSymbols
        this.getInputWidth = function() { return _InputWidth; };
        //noinspection JSUnusedGlobalSymbols
        this.setInputWidth = function(/** VRS.InputWidth */ value) { _InputWidth = value; };
        //endregion

        //region -- getValue, setValue, saveState
        /**
         * Calls the getValue method and returns the result.
         * @returns {*}
         */
        this.getValue = function()
        {
            return that.getGetValue()();
        };

        /**
         * Calls the setValue method with the value passed across.
         * @param {*} value
         */
        this.setValue = function(value)
        {
            that.getSetValue()(value);
        };

        /**
         * Calls the saveState method.
         */
        this.saveState = function()
        {
            that.getSaveState()();
        };
        //endregion

        //region -- getInputClass, applyInputClass
        /**
         * Returns the input class based on the setting for inputWidth.
         * @returns {string}
         */
        this.getInputClass = function()
        {
            return that.getInputWidth();
        };

        /**
         * Applies an input class to a jQuery element based on the setting for inputWidth.
         * @param jqElement
         */
        this.applyInputClass = function(jqElement)
        {
            var inputClass = that.getInputClass();
            if(inputClass && jqElement) jqElement.addClass(inputClass);
        };
        //endregion

        //region -- hookEvents, unhookEvents
        /**
         * Calls the hookChanged event hooker and records the result.
         * @param callback
         * @param forceThis
         */
        this.hookEvents = function(callback, forceThis)
        {
            if(_HookChanged) _ChangedHookResult = _HookChanged(callback, forceThis);
        };

        /**
         * Calls the unhookChanged event hooker with the result from the previous hookChanged call.
         */
        this.unhookEvents = function()
        {
            if(_Unhook && _ChangedHookResult) {
                _Unhook(_ChangedHookResult);
                _ChangedHookResult = undefined;
            }
        };
        //endregion
    };
    //endregion

    //region 'TypeSettings' objects - Option<Type-Of-Control-Here>Settings
    //region OptionButtonSettings
    /**
     * The settings that can be associated with a button control.
     * @param {object}  settings               Is actually optional but if you declare it as optional then WebStorm7 gets its knickers in a twist.
     * @param {string} [settings.icon]          The name of the jQuery UI icon to display on the button without the leading 'ui-icon-'.
     * @param {string} [settings.primaryIcon]   The name of the jQuery UI icon to display on the button when it isn't pressed, without the leading 'ui-icon-'.
     * @param {string} [settings.secondaryIcon] The name of the jQuery UI icon to display on the button when it is pressed, without the leading 'ui-icon'.
     * @param {bool}   [settings.showText]      True if the text is to be displayed, false if only the icon is to be displayed.
     * @constructor
     */
    VRS.OptionButtonSettings = function(settings)
    {
        if(!settings) settings = {};

        var _Dispatcher = new VRS.EventHandler({
            name: 'VRS.OptionButtonSettings'
        });
        var _Events = {
            enabledChanged: 'enabledChanged'
        };

        /** @type {string=} */
        this.primaryIcon = settings.primaryIcon || settings.icon;

        /** @type {string=} */
        this.secondaryIcon = settings.secondaryIcon;

        /** @type {boolean=} */
        this.showText = settings.showText;

        var _Enabled = true;
        this.getEnabled = function() { return _Enabled; };
        this.setEnabled = function(/** bool */ value) {
            if(value !== _Enabled) {
                _Enabled = value;
                _Dispatcher.raise(_Events.enabledChanged);
            }
        };

        this.hookEnabledChanged = function(/** function() */ callback, /** Object */ forceThis) { return _Dispatcher.hook(_Events.enabledChanged, callback, forceThis); };
        this.unhook = function(/** Object */ hookResult) { _Dispatcher.unhook(hookResult); };
    };
    //endregion

    //region OptionComboBoxSettings
    /**
     * The settings that can be associated with a combo-box control.
     * @param {object}               settings           Is actually optional but if you declare it as optional then WebStorm7 gets its knickers in a twist.
     * @param {VRS.ValueText[]}     [settings.values]   The values to display in the drop-down. text values are displayed as-is, textKey is an index into VRS.$$ or a function that returns the translated text.
     * @param {function(?)}         [settings.changed]  Called whenever the combo box value changes. Gets passed the current value of the combo box.
     * @constructor
     */
    VRS.OptionComboBoxSettings = function(settings)
    {
        if(!settings) settings = {};

        /** @type {Array.<VRS.ValueText>} */
        this.values = settings.values || [];
        /** @type {function(?)=} */
        this.changed = settings.changed || $.noop;
    };
    //endregion

    //region OptionDateSettings
    /**
     * The settings that can be associated with a date control.
     * @param {Object}  settings                        Is actually optional but if you declare it as optional then WebStorm7 gets in a huff.
     * @param {Date=}   settings.defaultDate            The default date to display in the control. Defaults to today.
     * @param {Date=}   settings.minDate                The earliest date that can be selected by the user. Leave undefined when there is no minimum date.
     * @param {Date=}   settings.maxDate                The latest date that can be selected by the user. Leave undefined when there is no maximum date.
     * @constructor
     */
    VRS.OptionDateSettings = function(settings)
    {
        if(!settings) settings = {};

        this.defaultDate = settings.defaultDate || null;
        this.minDate = settings.minDate || null;
        this.maxDate = settings.maxDate || null;
    };
    //endregion

    //region OptionLabelSettings
    /**
     * The settings that can be associated with a label control.
     * @param {object}           settings                Is actually optional but if you declare it as optional then WebStorm7 gets its knickers in a twist.
     * @param {VRS.LabelWidth}  [settings.labelWidth]    The width to assign to the label.
     * @constructor
     */
    VRS.OptionLabelSettings = function(settings)
    {
        if(!settings) settings = {};

        /** @type {VRS.LabelWidth=} */
        this.labelWidth = settings.labelWidth !== undefined ? settings.labelWidth : VRS.LabelWidth.Auto;
    };
    //endregion

    //region OptionNumericTypeSettings
    /**
     * The settings that can be associated with a number editor control.
     * @param {object}   settings               Is actually optional but if you declare it as optional then WebStorm7 gets its knickers in a twist.
     * @param {number}  [settings.min]          The smallest value that the user can enter.
     * @param {number}  [settings.max]          The largest value that the user can enter.
     * @param {number}  [settings.decimals]     The number of decimal places that will be shown to the user.
     * @param {number}  [settings.step]         The increment to use when automatically stepping values up or down.
     * @constructor
     */
    VRS.OptionNumericTypeSettings = function(settings)
    {
        if(!settings) settings = {};

        /** @type {number=} */
        this.min = settings.min;

        /** @type {number=} */
        this.max = settings.max;

        /** @type {number=} */
        this.decimals = settings.decimals;

        /** @type {number} */
        this.step = settings.step || 1;
    };
    //endregion

    //region OptionOrderedSubsetTypeSettings
    /**
     * The settings that can be associated with an editor that lets users choose a subset from a set of values.
     * @param {object}              settings                       Is actually optional but if you declare it as such then WebStorm7 gets its knickers in a twist.
     * @param {VRS.ValueText[]}    [settings.values]               The full set of values that the user can choose from.
     * @param {bool}               [settings.keepValuesSorted]     True if the unused values are to be sorted when displayed to the user.
     * @constructor
     */
    VRS.OptionOrderedSubsetTypeSettings = function(settings)
    {
        if(!settings) settings = {};

        /** @type {Array} */
        this.values = settings.values || [];

        /** @type {boolean} */
        this.keepValuesSorted = settings.keepValuesSorted || false;
    };
    //endregion

    //region OptionPaneListTypeSettings
    /**
     * The settings that can be associated with an editor for an array of objects, where each object is edited via its own option pane.
     * @param {object}                               settings                           Is actually optional but if you declare it as optional then WebStorm7 gets its knickers in a twist.
     * @param {VRS.OptionPane[]}                    [settings.panes]                    The option panes to show to the user.
     * @param {number}                              [settings.maxPanes]                 The maximum number of panes (and by extension, items in the array being edited) that can be displayed / edited. Undefined or -1 indicates there is no maximum.
     * @param {VRS.OptionPane}                      [settings.addPane]                  An option pane containing the controls that can be used to add a new pane (and therefore a new array item).
     * @param {bool}                                [settings.suppressRemoveButton]     True if the UI to remove individual items is not shown to the user. Defaults to false.
     * @param {function(bool, jQuery)}              [settings.refreshAddControls]       An optional method that it called to disable the add controls. Passed true if the user has reached maximum number of panes. Also passed the jQuery element of the parent of the add pane. If not supplied then all input and button controls in the add pane are disabled or enabled.
     * @constructor
     */
    VRS.OptionPaneListTypeSettings = function(settings)
    {
        if(!settings) settings = {};

        //region -- Fields
        var that = this;
        var _Dispatcher = new VRS.EventHandler({
            name: 'VRS.OptionPaneListTypeSettings'
        });
        /** @enum {string} @readonly @private */
        var _Events = {
            paneAdded: 'paneAdded',
            paneRemoved: 'paneRemoved',
            maxPanesChanged: 'maxPanesChanged'
        };
        //endregion

        //region -- Events
        /**
         * Raised when the user adds a new pane.
         * @param {function(VRS.OptionPane, number)}    callback    Passed the new pane and its index in the array. The listener is expected to add the associated array item.
         * @param {object}                              forceThis   The object to use as 'this' when calling the callback.
         * @returns {object}
         */
        this.hookPaneAdded = function(callback, forceThis) { return _Dispatcher.hook(_Events.paneAdded, callback, forceThis); };

        /**
         * Raised when the user removes an existing pane.
         * @param {function(VRS.OptionPane, number)}    callback    Passed the pane that was removed and its index. The listener is expected to remove the associated array item.
         * @param {object}                              forceThis   The object to use as 'this' when calling the callback.
         * @returns {object}
         */
        this.hookPaneRemoved = function(callback, forceThis) { return _Dispatcher.hook(_Events.paneRemoved, callback, forceThis); };

        /**
         * Raised after the maximum number of panes has changed.
         * @param {function()}  callback
         * @param {object}      forceThis
         * @returns {{}}
         */
        this.hookMaxPanesChanged = function(callback, forceThis) { return _Dispatcher.hook(_Events.maxPanesChanged, callback, forceThis); };

        /**
         * Unhooks a previously hooked event.
         * @param {object} hookResult
         */
        this.unhook = function(hookResult) { _Dispatcher.unhook(hookResult); };

        function onPaneAdded(/** VRS.OptionPane */ pane, /** number */ index)   { _Dispatcher.raise(_Events.paneAdded, [ pane, index ]); }
        function onPaneRemoved(/** VRS.OptionPane */ pane, /** number */ index) { _Dispatcher.raise(_Events.paneRemoved, [ pane, index]); }
        function onMaxPanesChanged()                                            { _Dispatcher.raise(_Events.maxPanesChanged); }
        //endregion

        //region -- Properties
        /** @type {number} @private */
        var _MaxPanes = settings.maxPanes !== undefined ? settings.maxPanes : -1;
        this.getMaxPanes = function() { return _MaxPanes; };
        //noinspection JSUnusedGlobalSymbols
        this.setMaxPanes = function(/** number */ value) {
            if(value !== _MaxPanes) {
                _MaxPanes = value;
                trimExcessPanes();
                onMaxPanesChanged();
            }
        };

        /** @type {Array.<VRS.OptionPane>} @private */
        var _Panes = settings.panes || [];
        /** @returns {Array.<VRS.OptionPane>} */
        this.getPanes = function() { return _Panes; };

        /** @type {VRS.OptionPane} @private */
        var _AddPane = settings.addPane || null;
        this.getAddPane = function() { return _AddPane; };
        this.setAddPane = function(/** VRS.OptionPane */ addPane) { _AddPane = addPane; };

        /** @type {boolean} @private */
        var _SuppressRemoveButton = undefined === settings.suppressRemoveButton ? false : settings.suppressRemoveButton;
        this.getSuppressRemoveButton = function() { return _SuppressRemoveButton; };
        //noinspection JSUnusedGlobalSymbols
        this.setSuppressRemoveButton = function(/** boolean */ value) { _SuppressRemoveButton = value; };

        /** @type {function(bool)} */
        var _RefreshAddControls = settings.refreshAddControls || function(/** bool */ disabled, /** jQuery */ addParentJQ) {
            $(':input', addParentJQ).prop('disabled', disabled);
            $(':button', addParentJQ).button('option', 'disabled', disabled);
        };
        this.getRefreshAddControls = function() { return _RefreshAddControls; };
        this.setRefreshAddControls = function(/** function(bool) */ value) { _RefreshAddControls = value; };
        //endregion

        //region -- addPane, removePane, trimExcessPanes, findPaneIndex, removePaneAt
        /**
         * Adds a new pane to the array at the index specified.
         * @param {VRS.OptionPane} pane
         * @param {number} [index]
         */
        this.addPane = function(pane, index)
        {
            if(index !== undefined) {
                _Panes.splice(index, 0, pane);
                onPaneAdded(pane, index);
            } else {
                _Panes.push(pane);
                onPaneAdded(pane, _Panes.length - 1);
            }
        };

        /**
         * Removes the pane from the array.
         * @param {VRS.OptionPane} pane
         */
        this.removePane = function(pane)
        {
            var index = findPaneIndex(pane);
            if(index === -1) throw 'Cannot find the pane to remove';
            this.removePaneAt(index);
        };

        /**
         * Removes panes until the count of panes no longer exceeds the maximum allowed.
         */
        function trimExcessPanes()
        {
            if(_MaxPanes !== -1) {
                while(_MaxPanes > _Panes.length) {
                    that.removePane(_Panes[_Panes.length - 1]);
                }
            }
        }

        /**
         * Returns the index of the pane in the array or -1 if the pane is not in the arrary.
         * @param {VRS.OptionPane} pane
         * @returns {number}
         */
        function findPaneIndex(pane)
        {
            var result = -1;
            var length = _Panes.length;
            for(var i = 0;i < length;++i) {
                if(_Panes[i] === pane) {
                    result = i;
                    break;
                }
            }

            return result;
        }

        /**
         * Removes the pane at the index specified.
         * @param {number} index
         */
        this.removePaneAt = function(index)
        {
            var pane = _Panes[index];
            _Panes.splice(index, 1);
            pane.dispose();
            onPaneRemoved(pane, index);
        };
        //endregion
    };
    //endregion

    //region OptionRadioButtonTypeSettings
    /**
     * The settings that can be associated with a radio button group control.
     * @param {object}               settings           Is actually optional but if you declare it as optional then WebStorm7 gets its knickers in a twist.
     * @param {VRS.ValueText[]}     [settings.values]   The values for each of the radio buttons to show to the user.
     * @constructor
     */
    VRS.OptionRadioButtonTypeSettings = function(settings)
    {
        if(!settings) settings = {};

        /** @type {VRS.ValueText[]} */
        this.values = settings.values || [];
    };
    //endregion

    //region OptionTextBoxTypeSettings
    /**
     * The settings that can be associated with an editor that can display and edit strings.
     * @param {object}   settings                   Is actually optional but if you declare it as optional then WebStorm7 gets its knickers in a twist.
     * @param {bool}    [settings.upperCase]        True if the text is to be forced into upper-case.
     * @param {bool}    [settings.lowerCase]        True if the text is to be forced into lower-case.
     * @param {number}  [settings.maxLength]        The maximum number of characters allowed in the box.
     * @constructor
     */
    VRS.OptionTextBoxTypeSettings = function(settings)
    {
        if(!settings) settings = {};

        /** @type {boolean} */
        this.upperCase = settings.upperCase || false;

        /** @type {boolean} */
        this.lowerCase = settings.lowerCase || false;

        /** @type {number} */
        this.maxLength = settings.maxLength;
    };
    //endregion
    //endregion

    //region OptionPane
    /**
     * Describes a collection of fields to present together on a page in the configuration UI.
     * @param {object}                          settings
     * @param {string}                          settings.name           The name of the pane.
     * @param {string|function():string}       [settings.titleKey]      The index into VRS.$$ or a method that returns a string. Used as the title for the pane on the page.
     * @param {number}                         [settings.displayOrder]  The relative display order of the pane. Lower display orders appear before higher display orders.
     * @param {VRS.OptionField[]}              [settings.fields]        The initial set of fields on the pane.
     * @param {function()}                     [settings.dispose]       A method to call when the pane is disposed of.
     * @constructor
     */
    VRS.OptionPane = function(settings)
    {
        if(!settings) throw 'You must supply settings';
        if(!settings.name) throw 'You must supply a name for the pane';

        //region -- Fields
        /**
         * The array of fields that the pane contains.
         * @type {Array.<VRS.OptionField>}
         * @private
         */
        var _OptionFields = [];

        /**
         * A value that is incremented any time a field is added to or removed from the pane.
         * @type {number}
         * @private
         */
        var _Generation = 0;
        //endregion

        //region -- Properties
        /** @type {string} @private */
        var _Name = settings.name;
        this.getName = function() { return _Name; };

        /** @type {string|function():string} * @private */
        var _TitleKey = settings.titleKey;
        this.getTitleKey = function()       { return _TitleKey; };
        //noinspection JSUnusedGlobalSymbols
        this.setTitleKey = function(/** string|function():string */ value)  { _TitleKey = value; };

        /** @type {number} @private */
        var _DisplayOrder = settings.displayOrder;
        this.getDisplayOrder = function()       { return _DisplayOrder; };
        //noinspection JSUnusedGlobalSymbols
        this.setDisplayOrder = function(/** number */ value)  { if(!isNaN(value)) _DisplayOrder = value; };

        this.getFieldCount = function() { return _OptionFields.length; };
        this.getField = function(idx)   { return _OptionFields[idx]; };
        //endregion

        //region -- dispose
        /**
         * Disposes of the pane.
         */
        this.dispose = settings.dispose || $.noop;
        //endregion

        //region -- addField, removeFieldByName, foreachField
        /**
         * Adds a field to the pane. Throws if the field already exists.
         * @param {VRS.OptionField} optionField
         */
        this.addField = function(optionField)
        {
            /// <summary>Adds the field to the list of option fields to display to the user.</summary>
            var existingIndex = findIndexByName(optionField.getName());
            if(existingIndex !== -1) throw 'There is already a field in this pane called ' + optionField.getName();
            _OptionFields.push(optionField);
            ++_Generation;
        };

        /**
         * Removes a field from the pane by its name. Throws if the field does not exist.
         * @param {string} optionFieldName
         */
        this.removeFieldByName = function(optionFieldName)
        {
            var index = findIndexByName(optionFieldName);
            if(index === -1) throw 'Cannot remove option field ' + optionFieldName + ', it does not exist.';
            _OptionFields.splice(index, 1);
            ++_Generation;
        };

        /**
         * Passes every field on the pane to a callback method. The callback is not allowed to add or remove fields from the pane.
         * @param {function(VRS.OptionField)} callback
         */
        this.foreachField = function(callback)
        {
            /// <summary>Loops through each field in the order that it was added to the pane and passes it to the callback.
            var length = _OptionFields.length;
            //noinspection UnnecessaryLocalVariableJS
            var generation = _Generation;       // <-- not redundant, need to ensure that callback doesn't mess with fields
            for(var i = 0;i < length;++i) {
                callback(_OptionFields[i]);
                if(_Generation !== generation) throw 'Cannot continue to iterate through the fields after the collection has been modified';
            }
        };

        /**
         * Returns the index of the named field or -1 if there is no field with that name.
         * @param {string} optionFieldName
         * @returns {number}
         */
        function findIndexByName(optionFieldName)
        {
            var result = -1;
            $.each(_OptionFields, function(idx, val) {
                var breakLoop = val.getName() === optionFieldName;
                if(breakLoop) result = idx;
                return !breakLoop;
            });

            return result;
        }
        //endregion

        //region -- Initialisation
        if(settings.fields) {
            for(var i = 0;i < settings.fields.length;++i) this.addField(settings.fields[i]);
        }
        //endregion
    };
    //endregion

    //region OptionPage
    /**
     * Describes a page of panes in the configuration UI.
     * @param {object}              settings
     * @param {string}              settings.name           The unique name of the page.
     * @param {number}             [settings.displayOrder]  The relative order of the page within a group of pages. Pages with a lower display order come before those with a higher order.
     * @param {VRS.OptionPane[]}   [settings.panes]         The initial set of panes.
     * @constructor
     */
    VRS.OptionPage = function(settings)
    {
        if(!settings) throw 'You must supply settings';
        if(!settings.name) throw 'You must give the page a name';

        //region -- Fields
        /**
         * The panes held by the page.
         * @type {Array.<VRS.OptionPane>}
         * @private
         */
        var _OptionPanes = [];

        /**
         * A value that is incremented every time a pane is added to or removed from the page.
         * @type {number}
         * @private
         */
        var _Generation = 0;

        /**
         * The generation number as-at the last time the page was sorted. Used to optimise the sort method.
         * @type {number}
         * @private
         */
        var _SortGeneration = -1;
        //endregion

        //region -- Properties
        /** @type {string} @private */
        var _Name = settings.name;
        this.getName = function()        { return _Name; };
        //noinspection JSUnusedGlobalSymbols
        this.setName = function(/** string */ value)   { _Name = value; };

        /** @type {string} @private */
        var _TitleKey = settings.titleKey;
        this.getTitleKey = function()       { return _TitleKey; };
        //noinspection JSUnusedGlobalSymbols
        this.setTitleKey = function(/** string */ value)  { _TitleKey = value; };

        /** @type {number} @private */
        var _DisplayOrder = settings.displayOrder;
        this.getDisplayOrder = function()       { return _DisplayOrder; };
        //noinspection JSUnusedGlobalSymbols
        this.setDisplayOrder = function(/** number */ value)  { if(!isNaN(value)) _DisplayOrder = value; };
        //endregion

        //region -- addPane, removePaneByName, foreachPane
        /**
         * Adds a pane to the page. If there is already a pane on the page with the same name then an exception is thrown.
         * @param {VRS.OptionPane|Array.<VRS.OptionPane>} optionPane
         */
        this.addPane = function(optionPane)
        {
            if(!(optionPane instanceof VRS.OptionPane)) {
                var length = optionPane.length;
                for(var i = 0;i < length;++i) {
                    this.addPane(optionPane[i]);
                }
            } else {
                var index = findIndexByName(optionPane.getName());
                if(index !== -1) throw 'There is already a pane on this page called ' + optionPane.getName();
                _OptionPanes.push(optionPane);
                ++_Generation;
            }
        };

        //noinspection JSUnusedGlobalSymbols
        /**
         * Removes the pane with the name passed across. If there is no pane on the page with this name then an exception is thrown.
         * @param {string} optionPaneName
         */
        this.removePaneByName = function(optionPaneName)
        {
            var index = findIndexByName(optionPaneName);
            if(index === -1) throw 'There is no pane called ' + optionPaneName;
            _OptionPanes.splice(index, 1);
            ++_Generation;
        };

        /**
         * Passes each pane on the page to a callback method, one at a time. The callback must not add or remove panes on the page.
         * @param {function(VRS.OptionPane)} callback
         */
        this.foreachPane = function(callback)
        {
            sortPanes();
            //noinspection UnnecessaryLocalVariableJS
            var generation = _Generation;           // <-- not redundant, need to ensure callback doesn't mess with panes
            var length = _OptionPanes.length;
            for(var i = 0;i < length;++i) {
                callback(_OptionPanes[i]);
                if(_Generation != generation) throw 'Cannot continue to iterate through the panes, they have been changed';
            }
        };

        /**
         * Returns the index of a pane or -1 if there is no pane with the name passed across.
         * @param {string} optionPaneName
         * @returns {number}
         */
        function findIndexByName(optionPaneName)
        {
            var result = -1;
            $.each(_OptionPanes, function(idx, val) {
                var foundMatch = val.getName() === optionPaneName;
                if(foundMatch) result = idx;
                return !foundMatch;
            });

            return result;
        }

        /**
         * Sorts panes into display order.
         */
        function sortPanes()
        {
            if(_SortGeneration !== _Generation) {
                _OptionPanes.sort(function(lhs, rhs) { return lhs.getDisplayOrder() - rhs.getDisplayOrder(); });
                _SortGeneration = _Generation;
            }
        }
        //endregion

        //region -- Initialisation
        if(settings.panes) {
            for(var i = 0;i < settings.panes.length;++i) this.addPane(settings.panes[i]);
        }
        //endregion
    };
    //endregion

    //region OptionControlTypeBroker
    /**
     * An object that can translate between control types and the jQuery UI plugin that handles the control.
     * @constructor
     */
    VRS.OptionControlTypeBroker = function()
    {
        //region -- Fields
        /**
         * The array of control types and the method to use to create the appropriate jQuery UI plugin.
         * @type {Object.<VRS.optionControlTypes, function(VRS.OptionField, jQuery, number=):jQuery>}
         * @private
         */
        var _ControlTypes = {};
        //endregion

        //region -- addControlTypeHandler, removeControlTypeHandler, controlTypeHasHandler
        /**
         * Adds a handler for a control type handler. Throws an exception if there is already a handler for the control type.
         * @param {string}                                              controlType         The VRS.optionControlTypes control type to register.
         * @param {function(VRS.OptionField, jQuery, number=):jQuery}   creatorCallback     A method that takes a parent element and creates the control for it.
         */
        this.addControlTypeHandler = function(controlType, creatorCallback)
        {
            if(_ControlTypes[controlType]) throw 'There is already a handler registered for ' + controlType + ' control types';
            _ControlTypes[controlType] = creatorCallback;
        };

        /**
         * Adds a handler for a control type, but only if one is not already registered. Does nothing if one is already registered.
         * @param {string}                                              controlType         The VRS.optionControlTypes control type to register.
         * @param {function(VRS.OptionField, jQuery, number=):jQuery}   creatorCallback     A method that takes a parent element and creates the control for it.
         */
        this.addControlTypeHandlerIfNotRegistered = function(controlType, creatorCallback)
        {
            if(!this.controlTypeHasHandler(controlType)) this.addControlTypeHandler(controlType, creatorCallback);
        };

        //noinspection JSUnusedGlobalSymbols
        /**
         * Removes the handler for a control type. Throws an exception if the control type does not have a handler registered.
         * @param {string} controlType The VRS.optionControlTypes control type to remove.
         */
        this.removeControlTypeHandler = function(controlType)
        {
            if(!_ControlTypes[controlType]) throw 'There is no handler registered for ' + controlType + ' control types';
            delete _ControlTypes[controlType];
        };

        /**
         * Returns true if the control type has a handler registered for it, false if it does not.
         * @param {string} controlType The VRS.optionControlTypes control type to test for.
         * @returns {boolean}
         */
        this.controlTypeHasHandler = function(controlType)
        {
            return !!_ControlTypes[controlType];
        };

        /**
         * Creates the jQuery UI plugin that will manage the display and editing of an option field.
         * @param {VRS.OptionField} optionField     The field to edit.
         * @param {jQuery}          parentJQ        The parent element to attach the field editor to.
         * @param {number}         [rowFieldNumber] The optional row/field number to assign to the field.
         * @returns {jQuery}
         */
        this.createControlTypeHandler = function(optionField, parentJQ, rowFieldNumber)
        {
            var controlType = optionField.getControlType();
            var creator = _ControlTypes[controlType];
            if(!creator) throw 'There is no handler registered for ' + controlType + ' control types';

            return creator(optionField, parentJQ, rowFieldNumber);
        };
        //endregion
    };
    //endregion

    //region Pre-builts
    VRS.optionControlTypeBroker = new VRS.OptionControlTypeBroker();
    //endregion

    //region Standard VRS control types
    /**
     * An modifiable enumeration of the different control types. 3rd parties can add their own control types to this.
     * @enum {string}
     */
    VRS.optionControlTypes = VRS.optionControlTypes || {};
    VRS.optionControlTypes.button = 'vrsButton';
    VRS.optionControlTypes.checkBox = 'vrsCheckBox';
    VRS.optionControlTypes.colour = 'vrsColour';
    VRS.optionControlTypes.comboBox = 'vrsComboBox';
    VRS.optionControlTypes.date = 'vrsDate';
    VRS.optionControlTypes.label = 'vrsLabel';
    VRS.optionControlTypes.numeric = 'vrsNumeric';
    VRS.optionControlTypes.orderedSubset = 'vrsOrderedSubset';
    VRS.optionControlTypes.paneList = 'vrsPaneList';
    VRS.optionControlTypes.radioButton = 'vrsRadioButton';
    VRS.optionControlTypes.textBox = 'vrsTextBox';
    //endregion
}(window.VRS = window.VRS || {}, jQuery));
